/*****************************************************************************/
/* Implémentation relative à la Question 1 de l'examen                       */
/* - La représentation en format (K,L) est obtenue en tronquant les "float"  */
/* - Regarder les fonctions main et multiply (pour 1c).                      */

/*****************************************************************************/

#include <stdio.h>
#include <stdint.h>
#include <math.h>

/*****************************************************************************/

// Implicitement, le code ci-dessus fait l'hypothèse que K,L sont plus petits
// que dans float et que la multiplication se fait dans un entier de 32 bits,
// donc K <= 8 et L <= 15. D'ailleurs, certaines exemples dans main ont été
// choisis spécifiquement pour K=4, L=5.

#define K 4
#define L 5

#define BIAS ((1 << (K-1))-1)
#define MAX_EXP ((1 << K)-1)
#define MAX_MANT ((1 << L)-1)

#define FLOAT_BIAS 127

// un float décomposé
typedef struct {
	unsigned sign:1;
	unsigned exponent:8;
	unsigned mantissa:23;
} decomp_t;

// le format (K,L)
typedef struct {
	unsigned sign:1;
	unsigned exponent:K;
	unsigned mantissa:L;
} mini;

/*****************************************************************************/
/* affichage binaire                                                         */

void output_bits (uint32_t val, int k)
{
	while (k--) printf("%d",!!(val & (1<<k)));
}

void output_mini (mini m)
{
	output_bits(m.sign,1); putchar('.');
	output_bits(m.exponent,K); putchar('.');
	output_bits(m.mantissa,L);
}

/*****************************************************************************/
/* Étant donné un entier val (avec "from" bits), on prend les "top" bits les */
/* plus significatifs. Avec -DROUNDING, on fait un arrondi en plus. Si cet   */
/* arrondi déborde, on augmente l'exposant (non demandé pour l'examen).      */

#ifdef ROUNDING
uint32_t get_mantissa (uint32_t val, int length, int top, int* exponent)
{
	val = val & ((1 << length)-1);
	val += (1 << (length-top-1));
	if (val & (1 << length)) (*exponent)++;
	return (val >> (length-top)) & ((1<<top)-1);
}
#else
uint32_t get_mantissa (uint32_t val, int length, int top, int* exponent)
{
	return (val >> (length-top)) & ((1<<top)-1);
	
}
#endif

/*****************************************************************************/
/* constants 0, +∞, -∞                                                       */

mini mini_zero () { return (mini) { 0,0,0 }; }
mini mini_infty (int sign) { return (mini) { sign,MAX_EXP,0 }; }

/*****************************************************************************/
/* conversion entre float et format (K,L), pour affichage seulement,         */
/* pour affichage seulement, sous l'hypothèse que K<=8 et L<=23              */

mini to_mini (float f)
{
	decomp_t d;
	mini m;

	uint32_t fi = *(uint32_t*)&f;
	d.sign = fi >> 31;
	d.exponent = (fi >> 23) & 0xff;
	d.mantissa = fi & 0x7fffff;

	if (d.exponent == 0 && d.mantissa == 0) return mini_zero();
	
	m.sign = d.sign;
	int tmp_exp = d.exponent + BIAS - FLOAT_BIAS;
	m.mantissa = get_mantissa(d.mantissa,23,L,&tmp_exp);

	if (tmp_exp < 1) return mini_zero();
	if (tmp_exp >= MAX_EXP) return mini_infty(d.sign);
	m.exponent = tmp_exp;
	
	return m;
}

float to_float (mini m)
{
	decomp_t d;
	if (m.exponent == 0) return 0.0;
	if (m.exponent == MAX_EXP && m.sign == 0) return INFINITY;
	if (m.exponent == MAX_EXP && m.sign == 1) return -INFINITY;

	d.sign = m.sign;
	d.exponent = m.exponent + FLOAT_BIAS - BIAS;
	d.mantissa = m.mantissa << (23-L);

	uint32_t fi = (d.sign << 31) | (d.exponent << 23) | d.mantissa;
	return *(float*)&fi;
}

/*****************************************************************************/
/* La multiplication entre deux flottantes en format (K,L). Cette solution   */
/* traite certains cas spéciaux, tout n'est pas demandé pour avoir les 100%  */
/* dans l'examen.                                                            */

mini multiply (mini x, mini y)
{
	mini z;

	// signe: ou exclusif
	z.sign = x.sign ^ y.sign;
	// exposant : addition, corriger pour la majoration
	int tmp_exp = x.exponent + y.exponent - BIAS;
	// mantisse : multiplication, prenant en compte le "1" implicit
	int tmp_mant = (MAX_MANT+1 + x.mantissa) * (MAX_MANT+1 + y.mantissa);

	// cas spéciaux : zéro, infinité
	if (x.exponent == 0 || y.exponent == 0) return mini_zero();
	if (x.exponent == MAX_EXP || y.exponent == MAX_EXP)
		return mini_infty(z.sign);

	// normaliser exposant et mantisse : le "1" la plus significatif sera
	// soit à la position 2L, soit à 2L+1 ; on extrait ensuit les L bits
	// les plus significatifs

	if (tmp_mant & (1 << (L+L+1))) { tmp_mant /= 2; tmp_exp += 1; }
	z.mantissa = get_mantissa(tmp_mant,L+L,L,&tmp_exp);
	z.exponent = tmp_exp;

	// cas spéciaux : débordement d'exposant
	if (tmp_exp < 1) return mini_zero();
	if (tmp_exp >= MAX_EXP) return mini_infty(z.sign);

	return z;
}

/*****************************************************************************/
/* Fonctions auxiliaires pour main.                                          */

void represent (float f)
{
	mini m = to_mini(f);

	printf("représentation de %9.4f = ",f); output_mini(m);
	printf(", valeur précise = %f\n", to_float(m));
}

// calculer le plus petit entier non représentable directement
void plus_petit ()
{
	int i = 1;
	while (i++)
	{
		float f = i;
		if (to_float(to_mini(f)) != f) break;
	}
	printf("le plus petit entier non représentable est %d\n\n",i);
}

void multiply_test (mini x, mini y)
{
	printf("x=%f, y=%f, ",to_float(x),to_float(y));
	printf("x*y (float) = %f, ",to_float(x)*to_float(y));
	printf("x*y (mini) = %f\n",to_float(multiply(x,y)));
}

/*****************************************************************************/

int main ()
{
	// Question 1(a)
	represent(12.0);
	represent(3.1415);
	represent(-34.5);	// dépend de l'arrondi

	// quelques extras
	putchar('\n');
	represent(0.0);
	represent(800);
	represent(-INFINITY);

	// Question 1(b)
	printf("\nplus grand entier:\n");
	mini m = { 0, MAX_EXP-1, MAX_MANT };
	printf(" en binaire: "); output_mini(m);
	printf("\n en décimal: %f\n\n",to_float(m));
	plus_petit();

	// Question 1(c), quelques exemples
	multiply_test(to_mini(12.0),to_mini(-3.0));
	multiply_test(to_mini(12.0),to_mini(3.1415));	// dépend de l'arrondi
	printf("\n");

	// Ces exemples demontrent les différents cas ou les arrondis
	// jouent un rôle (non demandé pour l'examen). Observer les
	// différences avec et sans -DROUNDING.
	mini a = (mini) { 0,BIAS,1 };
	mini b = (mini) { 0,BIAS,2 };
	mini c = (mini) { 0,BIAS,29 };
	mini d = (mini) { 0,BIAS,30 };

	multiply_test(a,c);
	multiply_test(a,d);
	multiply_test(b,c);
	multiply_test(b,d);
}
